"use strict";
const hrtime_now_1 = require("hrtime-now");
const typescript_ast_util_1 = require("typescript-ast-util");
const typescript_plugin_util_1 = require("typescript-plugin-util");
const PLUGIN_NAME = 'typescript-plugin-add-type';
const REFACTOR_ACTION_NAME = `${PLUGIN_NAME}-refactor-action`;
let ts;
let info;
const pluginDefinition = { getApplicableRefactors, getEditsForRefactor };
let target;
function getApplicableRefactors(fileName, positionOrRange) {
    const t0 = hrtime_now_1.now();
    const refactors = info.languageService.getApplicableRefactors(fileName, positionOrRange) || [];
    const program = info.languageService.getProgram();
    const sourceFile = program.getSourceFile(fileName);
    if (!sourceFile) {
        return refactors;
    }
    target = undefined;
    const node = typescript_ast_util_1.findChildContainingRange(sourceFile, typescript_ast_util_1.positionOrRangeToRange(positionOrRange));
    if (!node) {
        info.project.projectService.logger.info(`${PLUGIN_NAME} no getEditsForRefactor because findChildContainedRange undefined`);
        return refactors;
    }
    const noFilters = false;
    if (noFilters) {
        target = node;
    }
    else {
        const predicate = function isDeclarationWithNoDeclaredType(node, program) {
            return node && typescript_ast_util_1.isDeclaration(node) && !typescript_ast_util_1.hasDeclaredType(node, program); // declaration without a type declaration 
            /*hasName(node) && */
        };
        // target = predicate(node, program) && node
        let child;
        target = predicate(node, program) ? node :
            predicate(node.parent, program) ? node.parent :
                (child = typescript_ast_util_1.findChild(node, c => predicate(c, program), false)) ? child :
                    // (child = findChild(node.parent, c => predicate(c, program), false)) ? child : 
                    undefined;
    }
    if (!target) {
        info.project.projectService.logger.info(`${PLUGIN_NAME} no getEditsForRefactor because target not found undefined`);
        return refactors;
    }
    const name = node.name && target.name.getText() || '';
    refactors.push({
        name: `${PLUGIN_NAME}-refactor-info`,
        description: 'Add type',
        actions: [
            { name: REFACTOR_ACTION_NAME, description: 'Add type to ' + typescript_ast_util_1.getKindName(target.kind).replace(/Declaration/gi, '').toLowerCase() + ' "' + name + '"' }
        ],
    });
    info.project.projectService.logger.info(`${PLUGIN_NAME} getApplicableRefactors took ${hrtime_now_1.now() - t0}`);
    return refactors;
}
function getEditsForRefactor(fileName, formatOptions, positionOrRange, refactorName, actionName) {
    const t0 = hrtime_now_1.now();
    const refactors = info.languageService.getEditsForRefactor(fileName, formatOptions, positionOrRange, refactorName, actionName);
    if (actionName != REFACTOR_ACTION_NAME || !target) {
        return refactors;
    }
    const textChange = getFileTextChanges(target, info.languageService.getProgram());
    const refactorEditInfo = {
        edits: [{
                fileName,
                textChanges: [textChange],
            }],
        renameFilename: undefined,
        renameLocation: undefined,
    };
    info.project.projectService.logger.info(`${PLUGIN_NAME} getEditsForRefactor took ${hrtime_now_1.now() - t0}`);
    return refactorEditInfo;
}
/**
 * besides correcting type strings like getTypeStringForDeclaration, should this be in plugin-util ?
 * some cases to consider:
  // function f ()$TYPE {}
  // function()$TYPE {} // should not enter because of no name
  // method()$TYPE {}
  // class C { property1$TYPE = 2 }
  // class C {property1$TYPE = 'va'}
  // var a$TYPE = 2
  // function(param1$TYPE, param2){}
  // each((c) $TYPE => true)
  // each((c$TYPE)  => true)
 */
function getFileTextChanges(node, program) {
    let newText = ': ' + typescript_ast_util_1.getTypeStringForDeclarations(node, program);
    let start = (typescript_ast_util_1.findIdentifier(node) || node).getEnd(); // this will work for variable declaration and non-declaration nodes
    let length = 0;
    if (ts.isFunctionLike(node)) {
        // const idEnd = (node as ts.FunctionDeclaration).name ? (node as ts.FunctionDeclaration).name.end :
        // (node as ts.FunctionDeclaration).name
        start = node.parameters.end + 1;
        // length=3
    }
    else /*if(!isDeclaration(node) || ts.isVariableDeclaration(node) || ts.isPropertyDeclaration(node))*/ {
        //do nothing, default value for newText seems to be doing fine
    }
    return { newText, span: { start, length } };
}
module.exports = typescript_plugin_util_1.getPluginCreate(pluginDefinition, (modules, anInfo) => {
    ts = modules.typescript;
    info = anInfo;
    info.project.projectService.logger.info(`${PLUGIN_NAME} created`);
});
